#include <stdlib.h>

#include "struc.h"
#include "fonctions.h"


void ML_pixel_buffer(int x, int y, ML_Color color, char* buffer)
{
	if(x<0 || x>127 || y<0 || y>63) return;
	switch(color)
	{
		case BLACK:
			buffer[(y<<4)+(x>>3)] |= 128>>(x&7);
			break;
		case WHITE:
			buffer[(y<<4)+(x>>3)] &= ~(128>>(x&7));
			break;
		case XOR:
			buffer[(y<<4)+(x>>3)] ^= 128>>(x&7);
			break;
	}
}

void ML_bmp_or_buffer(unsigned char *bmp, int x, int y, int width, int height, char* buffer)
{
	unsigned short line;
	char shift, *screen, *p;
	int i, j, begin=0, end, real_width;

	p = (char*)&line;
	end=height;
	real_width=(width-1>>3<<3)+8;


	if(!bmp || x<0 || x>128-width || y<1-height || y>63 || width<1 || height<1) return;
	if(y < 0) begin = -y;
	if(y+height > 64) end = 64-y;
	shift = 8-(x&7);
	screen = buffer +(y+begin<<4)+(x>>3);
	for(i=begin ; i<end ; i++)
	{
		for(j=0 ; j<width-1>>3 ; j++)
		{
			line = bmp[i*(real_width>>3)+j]<<shift;
			screen[j] |= *p;
			if(shift!=8) screen[j+1] |= *(p+1);
		}
		line = (bmp[i*(real_width>>3)+j] & -1<<(real_width-width))<<shift;
		screen[j] |= *p;
		if(shift!=8 && x+real_width<129) screen[j+1] |= *(p+1);
		screen += 16;
	}
}

void ML_bmp_or_cl_buffer(unsigned char *bmp, int x, int y, int width, int height, char* buffer)
{
	unsigned short line;
	char shift, *screen, *p;
	int i, j, real_width, begin_x, end_x, begin_y, end_y;
	char bool1=1, bool2=1, bool3;
	if(!bmp || x<1-width || x>127 || y<1-height || y>63 || height<1 || width<1) return;
	p = (char*)&line;
	real_width = (width-1>>3<<3)+8;
	if(y < 0) begin_y = -y;
	else begin_y = 0;
	if(y+height > 64) end_y = 64-y;
	else end_y = height;
	shift = 8-(x&7);
	if(x<0)
	{
		begin_x = -x>>3;
		if(shift != 8) bool1 = 0;
	} else begin_x = 0;
	if(x+real_width > 128) end_x = 15-(x>>3), bool2 = 0;
	else end_x = real_width-1>>3;
	bool3 = (end_x == real_width-1>>3);
	screen = buffer+(y+begin_y<<4)+(x>>3);

	for(i=begin_y ; i<end_y ; i++)
	{
		if(begin_x < end_x)
		{
			line = bmp[i*(real_width>>3)+begin_x] << shift;
			if(bool1) screen[begin_x] |= *p;
			if(shift!=8) screen[begin_x+1] |= *(p+1);
			for(j=begin_x+1 ; j<end_x ; j++)
			{
				line = bmp[i*(real_width>>3)+j] << shift;
				screen[j] |= *p;
				if(shift!=8) screen[j+1] |= *(p+1);
			}
		}
		line = bmp[i*(real_width>>3)+end_x];
		if(bool3) line &= -1<<real_width-width;
		line <<= shift;
		if(begin_x < end_x || bool1) screen[end_x] |= *p;
		if(bool2) screen[end_x+1] |= *(p+1);
		screen += 16;
	}
}

void ML_horizontal_line_buffer(int y, int x1, int x2, ML_Color color, char* buffer)
{
    int i;
    char checker;
    char* vram;
	vram = buffer;
    if(y<0 || y>63 || (x1<0 && x2<0) || (x1>127 && x2>127)) return;
    if(x1 > x2)
    {
    	i = x1;
    	x1 = x2;
    	x2 = i;
    }
    if(x1 < 0) x1 = 0;
    if(x2 > 127) x2 = 127;
    switch(color)
    {
    	case BLACK:
			if(x1>>3 != x2>>3)
			{
				vram[(y<<4)+(x1>>3)] |= 255 >> (x1&7);
				vram[(y<<4)+(x2>>3)] |= 255 << 7-(x2&7);
				for(i=(x1>>3)+1 ; i<x2>>3 ; i++)
					vram[(y<<4) + i] = 255;
			}
			else vram[(y<<4)+(x1>>3)] |= (255>>(x1%8 + 7-x2%8))<<(7-(x2&7));
			break;
		case WHITE:
			if(x1>>3 != x2>>3)
			{
				vram[(y<<4)+(x1>>3)] &= 255 << 8-(x1&7);
				vram[(y<<4)+(x2>>3)] &= 255 >> 1+(x2&7);
				for(i=(x1>>3)+1 ; i<x2>>3 ; i++)
					vram[(y<<4) + i] = 0;
			}
			else vram[(y<<4)+(x1>>3)] &= (255<<8-(x1&7)) | (255>>1+(x2&7));
			break;
		case XOR:
			if(x1>>3 != x2>>3)
			{
				vram[(y<<4)+(x1>>3)] ^= 255 >> (x1&7);
				vram[(y<<4)+(x2>>3)] ^= 255 << 7-(x2&7);
				for(i=(x1>>3)+1 ; i<(x2>>3) ; i++)
					vram[(y<<4) + i] ^= 255;
			}
			else vram[(y<<4)+(x1>>3)] ^= (255>>((x1&7) + 7-(x2&7)))<<(7-(x2&7));
			break;
		case CHECKER:
			checker = (y&1 ? 85 : 170);
			if(x1>>3 != x2>>3)
			{
				vram[(y<<4)+(x1>>3)] &= 255 << 8-(x1&7);
				vram[(y<<4)+(x2>>3)] &= 255 >> 1+(x2&7);
				vram[(y<<4)+(x1>>3)] |= checker & 255>>(x1&7);
				vram[(y<<4)+(x2>>3)] |= checker & 255<<7-(x2&7);
				for(i=(x1>>3)+1 ; i<x2>>3 ; i++)
					vram[(y<<4) + i] = checker;
			}
			else
			{
				vram[(y<<4)+(x1>>3)] &= (255<<8-(x1&7)) | (255>>1+(x2&7));
				vram[(y<<4)+(x1>>3)] |= checker & (255>>(x1%8 + 7-x2%8))<<(7-(x2&7));
			}
			break;
    }
}

void ML_vertical_line_buffer(int x, int y1, int y2, ML_Color color, char* buffer)
{
    int i, j;
    char checker, byte;
	if(x<0 || x>127 || (y1<0 && y2<0) || (y1>63 && y2>63)) return;
	if(y1 > y2)
	{
		int tmp = y1;
		y1 = y2;
		y2 = tmp;
	}
	if(y1 < 0) y1 = 0;
	if(y2 > 63) y2 = 63;

	i = (y1<<4)+(x>>3);
	j = (y2<<4)+(x>>3);
	switch(color)
	{
		case BLACK:
			byte = 128>>(x&7);
			for( ; i<=j ; i+=16)
				buffer[i] |= byte;
			break;
		case WHITE:
			byte = ~(128>>(x&7));
			for( ; i<=j ; i+=16)
				buffer[i] &= byte;
			break;
		case XOR:
			byte = 128>>(x&7);
			for( ; i<=j ; i+=16)
				buffer[i] ^= byte;
			break;
		case CHECKER:
			byte = 128>>(x&7);
			checker = y1&1^x&1;
			for( ; i<=j ; i+=16)
			{
				if(checker) buffer[i] &= ~byte;
				else buffer[i] |= byte;
				checker = !checker;
			}
			break;
	}
}

void ML_rectangle_buffer(int x1, int y1, int x2, int y2, int border_width, ML_Color border_color, ML_Color fill_color, char* buffer)
{
	int i;
	if(x1 > x2)
	{
		i = x1;
		x1 = x2;
		x2 = i;
	}
	if(y1 > y2)
	{
		i = y1;
		y1 = y2;
		y2 = i;
	}
	if(border_width > (x2-x1)/2+1) border_width = (x2-x1)/2+1;
	if(border_width > (y2-y1)/2+1) border_width = (y2-y1)/2+1;
	if(border_color != TRANSPARENT && border_width > 0)
	{
		for(i=0 ; i<border_width ; i++)
		{
			ML_horizontal_line_buffer(y1+i, x1, x2, border_color, buffer);
			ML_horizontal_line_buffer(y2-i, x1, x2, border_color, buffer);
		}
		for(i=y1+border_width ; i<=y2-border_width ; i++)
		{
			ML_horizontal_line_buffer(i, x1, x1+border_width-1, border_color, buffer);
			ML_horizontal_line_buffer(i, x2-border_width+1, x2, border_color, buffer);
		}
	}
	if(fill_color != TRANSPARENT)
	{
		for(i=y1+border_width ; i<=y2-border_width ; i++)
			ML_horizontal_line_buffer(i, x1+border_width, x2-border_width, fill_color, buffer);
	}
}

void afficher_chaine(char* chaine, int x, int y, char* buffer)
{
	int i;
	unsigned int caractere[2] = {0};

    for(i=0; chaine[i] != '\0'; i++)//for(i=0; i<5; i++)
    {
		switch (chaine[i])
		{
			case 'A': caractere[0] = 0xF8282828 ; caractere[1] = 0xF8000000 ; break ;
			case 'B': caractere[0] = 0xF8A8A8A8 ; caractere[1] = 0x70000000 ; break ;
			case 'C': caractere[0] = 0xF8888888 ; caractere[1] = 0x88000000 ; break ;
			case 'D': caractere[0] = 0xF8888888 ; caractere[1] = 0x70000000 ; break ;
			case 'E': caractere[0] = 0xF8A8A8A8 ; caractere[1] = 0xA8000000 ; break ;
			case 'F': caractere[0] = 0xF8282828 ; caractere[1] = 0x28000000 ; break ;
			case 'G': caractere[0] = 0xF888A8A8 ; caractere[1] = 0xE8000000 ; break ;
			case 'H': caractere[0] = 0xF8202020 ; caractere[1] = 0xF8000000 ; break ;
			case 'I': caractere[0] = 0x8888F888 ; caractere[1] = 0x88000000 ; break ;
			case 'J': caractere[0] = 0x40808888 ; caractere[1] = 0x78000000 ; break ;
			case 'K': caractere[0] = 0xF8202050 ; caractere[1] = 0x88000000 ; break ;
			case 'L': caractere[0] = 0xF8808080 ; caractere[1] = 0x80000000 ; break ;
			case 'M': caractere[0] = 0xF8102010 ; caractere[1] = 0xF8000000 ; break ;
			case 'N': caractere[0] = 0xF8102040 ; caractere[1] = 0xF8000000 ; break ;
			case 'O': caractere[0] = 0xF8888888 ; caractere[1] = 0xF8000000 ; break ;
			case 'P': caractere[0] = 0xF8282828 ; caractere[1] = 0x38000000 ; break ;
			case 'Q': caractere[0] = 0xF88888C8 ; caractere[1] = 0xF8000000 ; break ;
			case 'R': caractere[0] = 0xF82828E8 ; caractere[1] = 0xB8000000 ; break ;
			case 'S': caractere[0] = 0xB8A8A8A8 ; caractere[1] = 0xE8000000 ; break ;
			case 'T': caractere[0] = 0x808F808  ; caractere[1] = 0x8000000  ; break ;
			case 'U': caractere[0] = 0xF8808080 ; caractere[1] = 0xF8000000 ; break ;
			case 'V': caractere[0] = 0x38408040 ; caractere[1] = 0x38000000 ; break ;
			case 'W': caractere[0] = 0x78807880 ; caractere[1] = 0x78000000 ; break ;
			case 'X': caractere[0] = 0x88502050 ; caractere[1] = 0x88000000 ; break ;
			case 'Y': caractere[0] = 0x1820C020 ; caractere[1] = 0x18000000 ; break ;
			case 'Z': caractere[0] = 0x88C8A898 ; caractere[1] = 0x88000000 ; break ;
			case '0': caractere[0] = 0xF88888   ; caractere[1] = 0xF8000000 ; break ;
			case '1': caractere[0] = 0x90F880   ; caractere[1] = 0          ; break ;
			case '2': caractere[0] = 0xE8A8A8A8 ; caractere[1] = 0xB8000000 ; break ;
			case '3': caractere[0] = 0x88A8A8A8 ; caractere[1] = 0xF8000000 ; break ;
			case '4': caractere[0] = 0x605048F8 ; caractere[1] = 0x40000000 ; break ;
			case '5': caractere[0] = 0xB8A8A8A8 ; caractere[1] = 0xE8000000 ; break ;
			case '6': caractere[0] = 0xF8A8A8A8 ; caractere[1] = 0xE8000000 ; break ;
			case '7': caractere[0] = 0x808C828  ; caractere[1] = 0x18000000 ; break ;
			case '8': caractere[0] = 0xF8A8A8A8 ; caractere[1] = 0xF8000000 ; break ;
			case '9': caractere[0] = 0xB8A8A8A8 ; caractere[1] = 0xF8000000 ; break ;
			default : caractere[0] = 0 ; caractere[1] = 0 ; break;
		}
		ML_bmp_or_buffer((unsigned char*)caractere, x, y + i*6, 5, 5, buffer);
	}
}

void afficher_nombres_buffer(int numb, int x, int y, char* buffer)
{
	int i, j, l=-1;
	int surface = 0xF888F800;

	if(numb < 0)
		return;
	else if(numb == 0)
	{
		ML_bmp_or_buffer((unsigned char*)&surface, x, y, 5, 3, buffer);
		return;
	}

	for(i=numb ; i ; i/=10)
        l++;
    for(i=numb, j=l; i ; i/=10, j--)
    {
		switch (i%10)
		{
			case 0: surface = 0xF888F800; break ;
			case 1: surface = 0x00F80000; break ;
			case 2: surface = 0xE8A8B800; break ;
			case 3: surface = 0xA8A8F800; break ;
			case 4: surface = 0x7840F800; break ;
			case 5: surface = 0xB8A8E800; break ;
			case 6: surface = 0xF8A8E800; break ;
			case 7: surface = 0x808F800;  break ;
			case 8: surface = 0xF8A8F800; break ;
			case 9: surface = 0xB8A8F800; break ;
			default :				break;

		}
		ML_bmp_or_buffer((unsigned char*)&surface, x, y+j*4, 5, 3, buffer);
	}

}

void ML_circle_buffer(int x, int y, int radius, ML_Color color, char* buffer)
{
	int plot_x, plot_y, d;

	if(radius < 0) return;
	plot_x = 0;
	plot_y = radius;
	d = 1 - radius;

	ML_pixel_buffer(x, y+plot_y, color, buffer);
	if(radius)
	{
		ML_pixel_buffer(x, y-plot_y, color, buffer);
		ML_pixel_buffer(x+plot_y, y, color, buffer);
		ML_pixel_buffer(x-plot_y, y, color, buffer);
	}
	while(plot_y > plot_x)
	{
		if(d < 0)
			d += 2*plot_x+3;
		else
		{
			d += 2*(plot_x-plot_y)+5;
			plot_y--;
		}
		plot_x++;
		if(plot_y >= plot_x)
		{
			ML_pixel_buffer(x+plot_x, y+plot_y, color, buffer);
			ML_pixel_buffer(x-plot_x, y+plot_y, color, buffer);
			ML_pixel_buffer(x+plot_x, y-plot_y, color, buffer);
			ML_pixel_buffer(x-plot_x, y-plot_y, color, buffer);
		}
		if(plot_y > plot_x)
		{
			ML_pixel_buffer(x+plot_y, y+plot_x, color, buffer);
			ML_pixel_buffer(x-plot_y, y+plot_x, color, buffer);
			ML_pixel_buffer(x+plot_y, y-plot_x, color, buffer);
			ML_pixel_buffer(x-plot_y, y-plot_x, color, buffer);
		}
	}
}

void ML_filled_circle_buffer(int x, int y, int radius, ML_Color color, char* buffer)
{
	int plot_x, plot_y, d;

	if(radius < 0) return;
	plot_x = 0;
	plot_y = radius;
	d = 1 - radius;

	ML_horizontal_line_buffer(y, x-plot_y, x+plot_y, color, buffer);
	while(plot_y > plot_x)
	{
		if(d < 0)
			d += 2*plot_x+3;
		else {
			d += 2*(plot_x-plot_y)+5;
			plot_y--;
			ML_horizontal_line_buffer(y+plot_y+1, x-plot_x, x+plot_x, color, buffer);
			ML_horizontal_line_buffer(y-plot_y-1, x-plot_x, x+plot_x, color, buffer);
		}
		plot_x++;
		if(plot_y >= plot_x)
		{
			ML_horizontal_line_buffer(y+plot_x, x-plot_y, x+plot_y, color, buffer);
			ML_horizontal_line_buffer(y-plot_x, x-plot_y, x+plot_y, color, buffer);
		}
	}
}

